"""Hatch build hook for dpgui."""
import os
import subprocess
from pathlib import Path
from shutil import copytree, rmtree
from typing import Any

import yaml
from hatchling.builders.hooks.plugin.interface import BuildHookInterface


def node_call(args, **kwargs):
    """Call node with subprocess."""
    return subprocess.call(["node", *args], **kwargs)


class CustomBuildHook(BuildHookInterface):
    """Custom build hook for dpgui."""

    PLUGIN_NAME = "build_node"

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        """Build the web app and copy to Python directory."""
        project_dir = Path(self.root)

        with open(project_dir / ".yarnrc.yml") as f:
            yarn_path = str(Path(yaml.load(f, Loader=yaml.Loader)["yarnPath"]))

        node_call([yarn_path], cwd=project_dir)
        node_call(
            [yarn_path, "build"],
            cwd=project_dir,
            env={
                **os.environ,
                "BASE_URL": "/",
                "VUE_APP_DPGUI_PYTHON": "1",
                "UV_USE_IO_URING": "0",
            },
        )

        bundle_html_path = project_dir / "dist"

        if not bundle_html_path.exists():
            raise RuntimeError("Failed to build the project with Yarn, please retry.")

        rmtree(project_dir / "dpgui" / "dist", ignore_errors=True)
        copytree(bundle_html_path, project_dir / "dpgui" / "dist")



"""Export all installed dargs inputs."""
import hashlib
import json

try:
    from importlib import metadata
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata

from dargs import ArgumentEncoder


def generate_input_json() -> str:
    """Generate JSON string for inputs of installed packages.

    Returns
    -------
    str
        JSON string for inputs of installed packages.
    """
    try:
        eps = metadata.entry_points(group="dpgui")
    except TypeError:
        eps = metadata.entry_points().get("dpgui", [])
    args = {}
    for ep in eps:
        obj = ep.load()()
        name = ep.name
        jdata = json.dumps(obj, cls=ArgumentEncoder, separators=(",", ":"))
        jhash = hashlib.sha256(jdata.encode("utf-8")).hexdigest()
        args[jhash] = {
            "name": name,
            "obj": obj,
        }
    return json.dumps(args, cls=ArgumentEncoder, separators=(",", ":"))



"""DP-GUI Python package."""
import logging

from .dargs import generate_input_json as generate_dpgui_templates
from .web import run as start_dpgui

logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.INFO)

__all__ = ["start_dpgui", "generate_dpgui_templates"]



"""Command line interface for the dpgui application."""

import argparse

from .web import run


def dpgui():
    """Command line interface for the dpgui application."""
    argparser = argparse.ArgumentParser(description="A web-based GUI for user inputs.")
    argparser.add_argument(
        "--port", type=int, default=6042, help="The port to serve DP-GUI on."
    )
    argparser.add_argument(
        "--bind_all",
        action="store_true",
        help=(
            "Serve on all public interfaces. This will expose your DP-GUI instance "
            "to the network on both IPv4 and IPv6 (where available)."
        ),
    )
    args = argparser.parse_args()
    run(port=args.port)



"""Web server for serving the web app."""

import logging
from pathlib import Path

from waitress import serve
from werkzeug.exceptions import HTTPException
from werkzeug.middleware.shared_data import SharedDataMiddleware
from werkzeug.routing import Map, Rule
from werkzeug.wrappers import Request, Response

from .dargs import generate_input_json

dist_dir = Path(__file__).parent / "dist"
log = logging.getLogger(__name__)


class App:
    """Web server for serving the web app."""

    def __init__(self) -> None:
        self.url_map = Map(
            [
                Rule("/", endpoint="app"),
                Rule("/<path:route>", endpoint="app"),
                Rule("/api/inputs", endpoint="inputs"),
            ]
        )
        self.wsgi_app = SharedDataMiddleware(
            self.wsgi_app,
            {
                "/": str(dist_dir),
            },
        )

    def dispatch_request(self, request):
        """Dispatch the request to the appropriate handler."""
        adapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint, values = adapter.match()
            return getattr(self, f"on_{endpoint}")(request, **values)
        except HTTPException as e:
            return e

    def on_app(self, request, **kwargs):
        """Serve the web app."""
        with open(dist_dir / "index.html") as f:
            html = f.read()
        return Response(html, mimetype="text/html")

    def on_inputs(self, request, **kwargs):
        """Serve the input JSON."""
        return Response(generate_input_json(), mimetype="application/json")

    def wsgi_app(self, environ, start_response):
        """WSGI application."""
        request = Request(environ)
        response = self.dispatch_request(request)
        return response(environ, start_response)

    def __call__(self, environ, start_response):
        """Call the WSGI application."""
        return self.wsgi_app(environ, start_response)


def run(port: int, bind_all: bool = False):
    """Run the web server."""
    app = App()
    if bind_all:
        listen = f"0.0.0.0:{port} [::]:{port}"
    else:
        listen = f"127.0.0.1:{port} [::1]:{port}"
    log.info(f"Open DP-GUI from http://localhost:{port}/")
    serve(app, listen=listen)
